from contextlib import contextmanager
import decorator
from greengen.generator import GreenletGenerator

_current_greengen = None


class NoCurrentGreengenException(SyntaxError):
    pass


def yield_(item):
    if _current_greengen is None:
        raise NoCurrentGreengenException('yield outside greengen')
    _current_greengen.yield_(item)


@decorator.decorator
def greengen(func, *args, **kwargs):
    curr_greengen = GreenletGenerator(func, *args, **kwargs)
    try:
        # Set the current greengen, so that calls to `yield_` will yield to this greengen.
        with _swap_current_greengen(curr_greengen) as prev_greengen:
            for item in curr_greengen:
                # We need to temporarily swap back to the previous greengen, so that nested greengens' calls to `yield_`
                # will yield to the correct greengen, and not the inner greengen.
                with _swap_current_greengen(prev_greengen):
                    yield item
    finally:
        # We have to call the `__del__` explicitly to prevent memory leaks caused by circular references.
        curr_greengen.__del__()


@contextmanager
def _swap_current_greengen(new_greengen):
    global _current_greengen
    backup = _current_greengen
    _current_greengen = new_greengen
    try:
        yield backup
    finally:
        _current_greengen = backup



from collections import namedtuple
from greenlet import greenlet


_ItemWrapper = namedtuple('_ItemWrapper', ['value'])


class GreenletGenerator(object):
    def __init__(self, func, *args, **kwargs):
        self.func = func
        self.args = args
        self.kwargs = kwargs
        self._next_item = None
        self._func_greenlet = None  # Will stay the same greenlet
        self._consume_greenlet = None  # Will change on each call to `next`

    def yield_(self, item):
        self._next_item = _ItemWrapper(item)
        self._consume_greenlet.switch()

    def _consume_next_item(self):
        if self._func_greenlet is None:
            self._func_greenlet = greenlet(self.func)
        self._func_greenlet.switch(*self.args, **self.kwargs)
        wrapped_item = self._next_item
        self._next_item = None
        return wrapped_item

    def next(self):
        self._consume_greenlet = greenlet(self._consume_next_item)
        wrapped_item = self._consume_greenlet.switch()
        if wrapped_item is None:
            raise StopIteration()
        return wrapped_item.value

    __next__ = next

    def __iter__(self):
        return self

    def __repr__(self):
        return '<GreenletGenerator: {}>'.format(self.func)

    def __del__(self):
        self._consume_greenlet = None
        self._func_greenlet = None



import os
from setuptools import setup, find_packages


BASE_DIR = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
README = open(os.path.join(BASE_DIR, 'README.md')).read()

setup(
    name='greengen',
    version='1.0.1',
    description='Nesting generators using greenlets',
    long_description=README,
    long_description_content_type="text/markdown",
    url="https://github.com/guylapid/greengen",
    author='guylapid',
    classifiers=[
        "Programming Language :: Python :: 2",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
    ],
    packages=find_packages(),
    include_package_data=True,
    install_requires=[
        'greenlet',
        'decorator',
    ],
)
